"""Use CCure CRUD operations to perform some common actions"""

from datetime import datetime, timezone
from typing import Optional

from acslib.base import (
    ACSRequestData,
    ACSRequestResponse,
    ACSRequestException,
    status,
)
from acslib.ccure.base import CcureACS
from acslib.ccure.connection import CcureConnection, ACSRequestMethod
from acslib.ccure.filters import (
    CcureFilter,
    ClearanceFilter,
    PersonnelFilter,
    NFUZZ,
)
from acslib.ccure.types import ObjectType, ImageType


class PersonnelAction(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = PersonnelFilter()
        self.type = ObjectType.PERSONNEL.complete

    def assign_clearances(self, personnel_id: int, clearance_ids: list[int]) -> ACSRequestResponse:
        """Assign clearances to a person"""
        clearance_assignment_properties = [
            {"PersonnelID": personnel_id, "ClearanceID": clearance_id}
            for clearance_id in clearance_ids
        ]
        return self.add_children(
            parent_type=ObjectType.PERSONNEL.complete,
            parent_id=personnel_id,
            child_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            child_configs=clearance_assignment_properties,
        )

    def revoke_clearances(self, personnel_id: int, clearance_ids: list[int]) -> ACSRequestResponse:
        """
        Revoke a person's clearances
        Two steps:
            1: Get the PersonnelClearancePair object IDs
            2: Remove those PersonnelClearancePair objects
        """

        # get PersonnelClearancePair object IDs
        clearance_query = " OR ".join(
            f"ClearanceID = {clearance_id}" for clearance_id in clearance_ids
        )
        search_filter = CcureFilter(display_properties=["PersonnelID", "ObjectID"])
        clearance_assignments = super().search(
            object_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            search_filter=search_filter,
            terms=[personnel_id],
            page_size=0,
            where_clause=f"PersonnelID = {personnel_id} AND ({clearance_query})",
        )
        assignment_ids = [assignment.get("ObjectID") for assignment in clearance_assignments]

        if assignment_ids:
            # remove PersonnelClearancePair objects
            return self.remove_children(
                parent_type=self.type,
                parent_id=personnel_id,
                child_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
                child_ids=assignment_ids,
            )

    def get_assigned_clearances(
        self, personnel_id: int, page_size=100, page_number=1
    ) -> list[dict]:
        """Get personnel/clearance pairs associated with the given person"""
        search_filter = CcureFilter(
            lookups={"PersonnelID": NFUZZ}, display_properties=["PersonnelID", "ClearanceID"]
        )
        return super().search(
            object_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            search_filter=search_filter,
            terms=[personnel_id],
            page_size=page_size,
            page_number=page_number,
        )

    def add_image(
        self, personnel_id: int, image: str, image_name: str = "", partition_id: int = 1
    ) -> ACSRequestResponse:
        """
        Set an image to a personnel object's PrimaryPortrait property
        - `image` is base-64 url-encoded.
        - `image_name` must be unique.
        - `partition_id` refers to the partition where the personnel object is stored.
        """
        if not image_name:
            timestamp = int(datetime.now(timezone.utc).timestamp())
            image_name = f"{personnel_id}_{timestamp}"
        image_properties = {
            "Name": image_name,
            "ParentId": personnel_id,
            "ImageType": ImageType.PORTRAIT.value,
            "PartitionID": partition_id,
            "Primary": True,  # this only adds primary portraits
            "Image": image,
        }
        return self.add_children(
            parent_type=ObjectType.PERSONNEL.complete,
            parent_id=personnel_id,
            child_type=ObjectType.IMAGE.complete,
            child_configs=[image_properties],
        )

    def get_image(self, personnel_id: int) -> Optional[str]:
        """
        Get the `PrimaryPortrait` property for the person with the given personnel ID.
        The returned image is a base-64 encoded string.
        """
        return self.get_property(self.type, personnel_id, "PrimaryPortrait")


class ClearanceAction(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = ClearanceFilter()
        self.type = ObjectType.CLEARANCE.complete

    def get_assignees(self, clearance_id: int, page_size=100, page_number=1) -> list[dict]:
        """Get clearance/personnel pairs belonging to the given clearance"""
        search_filter = CcureFilter(
            lookups={"ClearanceID": NFUZZ}, display_properties=["PersonnelID", "ClearanceID"]
        )
        return super().search(
            object_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            search_filter=search_filter,
            terms=[clearance_id],
            page_size=page_size,
            page_number=page_number,
        )


class DoorAction(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.type = ObjectType.DOOR.complete

    def lock(
        self,
        door_id: int,
        lock_time: Optional[datetime] = None,
        unlock_time: Optional[datetime] = None,
        priority: Optional[int] = None,
        source_name: str = "acslib",
    ):
        """
        Lock a door for a set period of time
        `lock_time`, `unlock_time`, and `priority` are optional.
        If there are multiple conflicting schedules, the schedule with the higher priority value
            will take precedence
        `source_name` refers to the client application making the request
        """
        if lock_time and unlock_time and lock_time > unlock_time:
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST,
                log_message="unlock_time must be after lock_time.",
            )
        TIME_FORMAT = "%m/%d/%Y %H:%M:%S"  # MM/DD/YYYY hh:mm:ss
        property_names = ["TargetType", "TargetID"]
        property_values = [ObjectType.ISTAR_DOOR.complete, door_id]
        if lock_time:
            property_names.append("StartTime")
            property_values.append(lock_time.strftime(TIME_FORMAT))
        if unlock_time:
            property_names.append("EndTime")
            property_values.append(unlock_time.strftime(TIME_FORMAT))
        if priority:
            property_names.append("Priority")
            property_values.append(priority)
        request_data = {
            "PropertyNames": property_names,
            "PropertyValues": property_values,
            "sourceName": source_name,
        }
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.ACTION,
                params={"actionTypeFullName": "SoftwareHouse.NextGen.Common.Actions.LockDoor"},
                data=self.connection.encode_data(request_data),
                headers=self.connection.base_headers | self.connection.header_for_form_data,
            ),
        )

    def unlock(
        self,
        door_id: int,
        unlock_time: Optional[datetime] = None,
        lock_time: Optional[datetime] = None,
        priority: Optional[int] = None,
        source_name: str = "acslib",
    ):
        """
        Unlock a door for a set period of time
        `unlock_time`, `lock_time`, and `priority` are optional.
        If there are multiple conflicting schedules, the schedule with the higher priority value
            will take precedence
        `source_name` refers to the client application making the request
        """
        if unlock_time and lock_time and unlock_time > lock_time:
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST,
                log_message="lock_time must be after unlock_time.",
            )
        TIME_FORMAT = "%m/%d/%Y %H:%M:%S"  # MM/DD/YYYY hh:mm:ss
        property_names = ["TargetType", "TargetID"]
        property_values = [ObjectType.ISTAR_DOOR.complete, door_id]
        if unlock_time:
            property_names.append("StartTime")
            property_values.append(unlock_time.strftime(TIME_FORMAT))
        if lock_time:
            property_names.append("EndTime")
            property_values.append(lock_time.strftime(TIME_FORMAT))
        if priority:
            property_names.append("Priority")
            property_values.append(priority)

        request_data = {
            "PropertyNames": property_names,
            "PropertyValues": property_values,
            "sourceName": source_name,
        }
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.ACTION,
                params={"actionTypeFullName": "SoftwareHouse.NextGen.Common.Actions.UnLockDoor"},
                data=self.connection.encode_data(request_data),
                headers=self.connection.base_headers | self.connection.header_for_form_data,
            ),
        )


class CcureAction:
    def __init__(self, connection: Optional[CcureConnection] = None):
        self.personnel = PersonnelAction(connection)
        self.clearance = ClearanceAction(connection)
        self.door = DoorAction(connection)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Optional

import requests
from requests import structures
from pydantic import BaseModel

from acslib.base import status
from acslib.base.config import ACSConfig


class ACSConnectionException(Exception):
    pass


class ACSRequestException(Exception):
    """
    Exception raised on a failed request, including exceptions
    from the requests module and 400+ status codes
    """

    def __init__(self, status_code: int, log_message: str):
        self.status_code = status_code
        self.message = log_message
        self.exception_name = "RequestException"

    def __str__(self):
        return f"{self.exception_name}: {self.status_code} {self.message}"

    def __reduce__(self):
        return (ACSRequestException, (self.status_code, self.message))


class ACSNotImplementedException(ACSRequestException):
    def __init__(self, log_message: str):
        self.status_code = status.HTTP_501_NOT_IMPLEMENTED
        self.message = log_message
        self.exception_name = "NotImplementedException"


class ACSRequestMethod(Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"


class ACSRequestResponse:
    """Successful queries from handle_request return this type of object"""

    def __init__(self, status_code: int, json: Any, headers: structures.CaseInsensitiveDict):
        self.status_code = status_code
        self.json = json
        self.headers = headers

    def count(self):
        return len(self.json)


class ACSRequestData(BaseModel):
    """Kwargs used in requests get/post/etc methods"""

    url: str
    # query params:
    params: Optional[dict] = None
    # body x-www-form-urlencoded data:
    data: Optional[dict | str] = None
    # body raw json:
    request_json: Optional[dict] = None
    headers: Optional[dict] = None


class ACSConnection(ABC):
    REQUEST_TYPE = {
        ACSRequestMethod.GET: requests.get,
        ACSRequestMethod.POST: requests.post,
        ACSRequestMethod.PUT: requests.put,
        ACSRequestMethod.DELETE: requests.delete,
        ACSRequestMethod.PATCH: requests.patch,
    }

    def __init__(self, **kwargs):
        self.config: ACSConfig = kwargs["config"]
        self.timeout = kwargs.get("timeout", self.config.timeout)
        self.response = None

    @abstractmethod
    def login(self):
        pass

    @abstractmethod
    def logout(self):
        pass

    def _make_request(self, requests_method: ACSRequestMethod, request_data_map: dict):
        if req := self.REQUEST_TYPE.get(requests_method):
            return req(**request_data_map)
        raise ACSConnectionException(f"Invalid request method: {requests_method}")

    def request(
        self, requests_method: ACSRequestMethod, request_data: ACSRequestData, timeout: float
    ) -> ACSRequestResponse:
        """
        Process requests to remote servers.
        Either return a response with the resulting status code, json data, and headers,
        or raise an exception with the appropriate status code

        Parameters:
            requests_method: A method from the requests module. requests.get, requests.post, etc
            request_data: Data used as kwargs for the requests_method

        Returns: An object with status_code, json, and headers attributes
        """

        try:
            request_data_map = request_data.model_dump()
            request_data_map["json"] = request_data_map.pop("request_json", None)
            request_data_map["data"] = request_data_map.get("data", {})
            request_data_map["timeout"] = timeout
            # remove request_data_map properties with None values
            request_data_map = {k: v for k, v in request_data_map.items() if v is not None}
            response = self._make_request(requests_method, request_data_map)
        except requests.HTTPError:
            # An HTTP error occurred.
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST,
                log_message="An error occurred with this request",
            )
        except requests.URLRequired:
            # A valid URL is required to make a request.
            raise ACSRequestException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                log_message="A valid URL wasn't provided for this request",
            )
        except requests.TooManyRedirects:
            # Too many redirects.
            raise ACSRequestException(
                status_code=status.HTTP_421_MISDIRECTED_REQUEST, log_message="Too many redirects"
            )
        except requests.ConnectTimeout:
            # The request timed out while trying to connect to the remote server.
            # Requests that produced this error are safe to retry.
            raise ACSRequestException(
                status_code=status.HTTP_504_GATEWAY_TIMEOUT,
                log_message=f"Unable to connect to remote server in {timeout} second(s)",
            )
        except requests.ReadTimeout:
            # The server did not send any data in the allotted amount of time.
            raise ACSRequestException(
                status_code=status.HTTP_504_GATEWAY_TIMEOUT,
                log_message=f"No response from remote server in {timeout} second(s)",
            )
        except requests.Timeout:
            # The request timed out.
            # Watching this error will catch both ConnectTimeout and ReadTimeout errors.
            raise ACSRequestException(
                status_code=status.HTTP_504_GATEWAY_TIMEOUT,
                log_message=f"Request took longer than {timeout} second(s)",
            )
        except requests.ConnectionError:
            # A Connection error occurred.
            raise ACSRequestException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                log_message="Could not connect to the remote host",
            )
        except requests.RequestException:
            # There was an ambiguous exception that occurred while handling your request.
            raise ACSRequestException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                log_message="An exception occurred while handling this request",
            )
        if response.status_code in range(200, 300):
            return ACSRequestResponse(
                status_code=response.status_code, json=response.json(), headers=response.headers
            )
        if response.status_code == status.HTTP_500_INTERNAL_SERVER_ERROR:
            raise ACSRequestException(
                status_code=status.HTTP_502_BAD_GATEWAY, log_message=response.text
            )
        raise ACSRequestException(status_code=response.status_code, log_message=response.text)
